/* $Id: qmail-remove.c,v 1.7 2002/07/17 21:01:15 josh Exp $ */

/*
 * This program is loosely based on the python script which
 * does a similar thing and can be found at:
 *
 * http://www.redwoodsoft.com/~dru/programs/
 *
 */

/*
 * NOTES: 
 *
 *     yanked directory is relative to queue directory unless an
 *     absolute path is specified (begins with /)
 *
 *     queue directory should be absolute path
 *
 *     tested on Linux and OpenBSD
 *
 */

/* TODO:
 *
 * 	cleanup the path generation routines
 *
 */

#include <sys/types.h>
#include <sys/stat.h>
#include <dirent.h>
#include <errno.h>
#include <fcntl.h>
#include <limits.h>
#include <regex.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <fts.h>

/* many linux fcntl.h's seem to be broken */
#ifndef O_NOFOLLOW
#define O_NOFOLLOW  0400000
#endif

/* prototypes */
void usage(void);
int search_file(const char *filename, const char *regex);
int remove_file(const char *filename);
char *read_file(const char *filename);
int find_files(char *path[], const char *pattern);
unsigned long digits(unsigned long num);
char * mk_nohashpath(char *queue, int inode_name);
char * mk_hashpath(char *queue, int inode_name);
char * mk_newpath(char *queue, int inode_name);

/* globals */
extern const char *__progname;
const char *default_queue = "/var/qmail/queue";
const char *default_pattern = ".*";
int regex_flags = 0, verbosity = 0, conf_split = 23, remove_files = 0;
char *yank_dir = "yanked";
const char cvsrid[] = "$Id: qmail-remove.c,v 1.7 2002/07/17 21:01:15 josh Exp $";
int queue_fd = -1;

/* queues i
 * 
 * NOTE: the first queue must be mess as it is used as a key
 *
 */
const char  *queues[] = {"mess", "local", "remote", "info", "intd", "todo", "bounce", NULL};

int main(int argc, char **argv)
{
    long int split = 0;
    int ch, matches;
    char *pattern = NULL;
    
    if (argc < 2)
        usage();

    while ((ch = getopt(argc, argv, "ehip:q:rs:vy:?")) != -1) {
        switch (ch) {
	  case 'e':
	    regex_flags |= REG_EXTENDED;
	  case 'i':
	    regex_flags |= REG_ICASE;
	    break;
          case 'p':
            pattern = optarg;
            break;
          case 'q':
            queue_fd = open(optarg, O_RDONLY);
            if(queue_fd < 0) {
                perror("open()");
                exit(1);
            }
            if(fchdir(queue_fd) == -1) {
                perror("fchdir()");
                exit(1);
            }
	    close(queue_fd);
            break;
          case 'r':
            remove_files = 1;
            break;
          case 's':
            split = strtoul(optarg, NULL, 10);
            if(split == ULONG_MAX) {
                fprintf(stderr, "invalid split value\n");
                exit(EXIT_FAILURE);
            }
            conf_split = split;
            break;
          case 'v':
            verbosity ++;
            break;
          case 'y':
            yank_dir = optarg;
            break;
          default:
	    usage();
 	    break;
        }
    }

    if(queue_fd == -1) {
        queue_fd = open(default_queue, O_RDONLY);
        if(queue_fd < 0) {
            fprintf(stderr, "no queue directory specified and %s doesn't exist\n", 
                    default_queue);
            exit(1);
        }
        if(fchdir(queue_fd) == -1) {
            perror("fchdir()");
            exit(1);
        }
	    close(queue_fd);
    }

    matches = find_files((char **)queues, (pattern ? pattern : default_pattern));
    if(matches >= 0) {
        fprintf(stderr, "%d file(s) match\n", matches);
    }

    exit(EXIT_SUCCESS);
}


void usage(void)
{
    fprintf(stderr, "%s [options]\n", __progname);
    fprintf(stderr, "  -e\t\tuse extended POSIX regular expressions\n");
    fprintf(stderr, "  -h, -?\tthis help message\n");
    fprintf(stderr, "  -i\t\tsearch case insensitively [default: case sensitive]\n");
    fprintf(stderr, "  -p <pattern>\tspecify the pattern to search for\n");
    fprintf(stderr, "  -q <queuedir>\tspecify the base qmail queue dir [default: /var/qmail/queue]\n");
    fprintf(stderr, "  -r\t\tactually remove files, without this we'll only print them\n");
    fprintf(stderr, "  -s <split>\tspecify your conf-split value if non-standard [default: 23]\n");
    fprintf(stderr, "  -v\t\tincrease verbosity (can be used more than once)\n");
    fprintf(stderr, "  -y <yankdir>\tdirectory to put files yanked from the queue [default: <queuedir>/yanked]\n"); 
    exit(EXIT_FAILURE);
}


char * read_file(const char *filename)
{
    off_t bytes;
    int fd;
    char *buff = NULL;
    struct stat fd_stat;

    if(filename == NULL)
        return NULL;

    fd = open(filename, (O_RDONLY | O_NOFOLLOW), 0);
    if(fd != -1) {
        if(fstat(fd, &fd_stat) == 0) {
            bytes = fd_stat.st_size;
            buff = malloc(bytes + 1);
            if(buff != NULL) {
                buff[bytes] = '\0';
                if(read(fd, buff, bytes) != bytes) {
                    fprintf(stderr, "%s: read() too short\n", __progname);
                    free(buff);
                    buff = NULL;
                }
            } else {
                perror("malloc()"); 
            }
        } else {
            perror("fstat()");
        }
        close(fd);
    }
    return buff;
}


int search_file(const char *filename, const char *pattern) 
{
    regex_t match_me;
    int err_code, match = 0; 
    char *file_inards = NULL, error_str[20];

    if(pattern == NULL)
        return (-1);
    if(filename == NULL)
        return (-1);

    file_inards = read_file(filename);
    if(file_inards != NULL) {
        err_code = regcomp(&match_me, pattern, regex_flags | REG_NOSUB);
        if(err_code == 0) {
            if(regexec(&match_me, file_inards, 0, NULL, 0) == 0)
                match = 1;
        } else {
            /* regex error */
            regerror(err_code, &match_me, error_str, 20); 
            fprintf(stderr, "regcomp(): %s\n", error_str);
        }
        regfree(&match_me);
        free(file_inards);
    }

    if(match == 1)
        return (0);

    return (-1);
}


int find_files (char *dir_list[], const char *pattern)
{
    FTS *fts;
    FTSENT *ftsp;
    char *error_str = NULL, *argv[2];
    int i = 0, tmp_fd = -1;

    argv[0] = dir_list[0];
    argv[1] = NULL;

    if ((fts = fts_open((char **)argv, FTS_PHYSICAL, NULL)) == NULL) {
        perror("fts_open");
        return -1;
    }

    errno = 0;
    while ((ftsp = fts_read(fts)) != NULL) {
        switch (ftsp->fts_info) {
            case FTS_F:
                printf("%s: ", ftsp->fts_accpath);
                if(search_file(ftsp->fts_accpath, pattern) == 0) {
                    printf("yes\n");
                    i++;
                    tmp_fd = open(".", O_RDONLY);
                    if((tmp_fd >= 0) && (remove_files == 1)) {
                        fchdir(queue_fd);
                        remove_file(ftsp->fts_name);
                        if(fchdir(tmp_fd) != 0)
			                perror("fchdir()");
                    }
		            if(tmp_fd >= 0) {
			            close(tmp_fd);
			            tmp_fd = -1;
		            }
                } else {
                    printf("no\n");
                }
                break;
            case FTS_DNR: /* couldn't read */
            case FTS_ERR: /* error */
            case FTS_NS:  /* no stat info */
                error_str = strerror(ftsp->fts_errno);
                fprintf(stderr, "fts_read() %s: %s\n", ftsp->fts_path, error_str);
                break;
            default:
                break;
        }
    }

    fts_close(fts);

    return i;
}


/*
 * remove_file()
 *
 *     Takes a filename and assumes it is the path to a qmail queue file
 * (named after its inode). It attempts to find it in all the queues and
 * move them to the global "yank_dir". It returns the inode number of the
 * file it removed from the queue or -1 on an error.
 */
int remove_file(const char *filename)
{
    int i, count = 0;
    unsigned long inode_num;
    char *my_name, *old_name = NULL, *new_name = NULL;
    struct stat statinfo;


    if(filename == NULL) {
        fprintf(stderr, "remove_file(): no filename\n");
        return -1;
    }
            
    my_name = strrchr(filename, '/');
    if(my_name == NULL)
        my_name = (char *)filename;
    else
        my_name++;

    inode_num = strtoul(my_name, NULL, 10);
    if((inode_num == ULONG_MAX) || (inode_num == 0)) {
        fprintf(stderr, "%s doesn't look like an inode number\n", my_name);
        return -1;
    }

    for(i = 0; (queues[i] != NULL); i++) {
        new_name = mk_newpath((char *)queues[i], inode_num);
        if(new_name == NULL) {
            fprintf(stderr, "remove_file(): unable to create new name\n");
	        return -1;
        }
        old_name = mk_hashpath((char *)queues[i], inode_num);
        if(old_name == NULL) {
	        free(new_name);
            fprintf(stderr, "remove_file(): unable to create old name\n");
	        return -1;
        }
        if(rename(old_name, new_name) == 0) {
            /* succeeded */
            fprintf(stderr, "moved %s to %s\n", old_name, new_name);
            count ++;
        } else {
            if(errno == ENOENT) {
	            if(old_name) {
                    if (verbosity >= 2) {
                        fprintf(stderr, "remove_file(%s): not a file\n", old_name);
                    } 
		            free(old_name);
		            old_name = NULL;
	            }
                old_name = mk_nohashpath((char *)queues[i], inode_num);
	            if(old_name == NULL) {
		            free(new_name);
		            return -1;
	            }
                if(stat(old_name, &statinfo) == -1) {
                    if (verbosity >= 2) {
                        fprintf(stderr, "remove_file(%s): no stat info\n", old_name);
                    }
                    continue;
                }
                if( !S_ISREG(statinfo.st_mode) ) {
                    if (verbosity >= 2) {
                        fprintf(stderr, "remove_file(%s): not a file\n", old_name);
                    }
                    continue;
                }
                if(rename(old_name, new_name) == 0) {
                    /* succeeded */
                    fprintf(stderr, "moved %s to %s\n", old_name, new_name);
                    count ++;
                } else {
                    if(errno != ENOENT) {
                        perror("rename()");
                    }
                }
            } else {
                /* failed but exists */
                perror("rename()");
            }
        }
    }

    /* garbage collection */
    if(old_name)
        free(old_name);
    if(new_name)
        free(new_name);

    return count;
}

char * mk_nohashpath(char *queue, int inode_name)
{
    int len = 0;
    char * old_name = NULL;

    if((queue == NULL) || (inode_name <= 0)) 
        return NULL;

    len = strlen(queue);
    len += digits(inode_name);
    len += 4;
    old_name = malloc(len);

    if(old_name) {
        snprintf(old_name, len, "%s/%u", queue, inode_name);
    /*    fprintf(stderr, "path: [%s][%u]\n", old_name, inode_name); */
        return old_name;
    } else {
        perror("malloc()");
        return NULL;
    }
}

char * mk_hashpath(char *queue, int inode_name)
{
    int len = 0, hash_num = 0;
    char * old_name = NULL;

    if((queue == NULL) || (inode_name <= 0)) 
        return NULL;
    
    hash_num = (inode_name % conf_split);

    len = strlen(queue);
    len += digits(hash_num);
    len += digits(inode_name);
    len += 4;
    old_name = malloc(len);

    if(old_name) {
        snprintf(old_name, len, "%s/%u/%u", queue, hash_num, inode_name);
        /* fprintf(stderr, "path: [%s][%u]\n", old_name, inode_name); */
        return old_name;
    } else {
        perror("malloc()");
        return NULL;
    }
}

char * mk_newpath(char *queue, int inode_name)
{
    int len = 0;
    char *new_name = NULL;

    if((queue == NULL) || (inode_name <= 0)) {
        fprintf(stderr, "mk_newpath(): invalid queue\n");
        return NULL;
    }

    len = strlen(queue);
    len += strlen(yank_dir);
    len += digits(inode_name);
    len += 4;
    new_name = malloc(len);

    if(new_name) {
        snprintf(new_name, len, "%s/%u.%s", yank_dir, inode_name, queue);
        return new_name;
    } else {
        perror("malloc()");
        return NULL;
    }
}

/*
 * digits()
 *
 *     Returns the number of digits needed to represent the number "num".
 *
 */
unsigned long digits(unsigned long num) 
{
    unsigned long i = 0;
    while(num >= 10) {
        num /= 10;
        i++;
    }
    i++;
    return(i);
}


